#include <jlog/jlog.hpp>
#include <fstream>
#include <iostream>
#include <chrono>

#ifdef WIN32
    #define NOMINMAX
    #include <windows.h>
#endif

namespace jlog
{
    static std::string default_logfile = "latest.log";

    void set_default_logfile(const std::string& filename)
    {
        default_logfile = filename;
    }


    std::string get_timestamp()
    {
        using namespace std::chrono;
        auto const timestamp = current_zone()->to_local(system_clock::now());
        auto dp = floor<days>(timestamp);
        year_month_day ymd{floor<days>(timestamp)};
        hh_mm_ss time{floor<milliseconds>(timestamp - dp)};
        auto y = ymd.year();
        auto m = ymd.month();
        auto d = ymd.day();
        auto h = time.hours();
        auto M = time.minutes();
        auto s = time.seconds();
        auto ms = time.subseconds();
        return std::format("{}-{}-{} {}:{}:{}.{}", y, m, d, h.count(), M.count(), s.count(), ms.count());
    }

    void log_to_console(const std::string& message)
    {
#ifdef WIN32
        mcolor::windowsSaneify();
#endif
        std::cout << message;
    }

    void log_to_file(const std::string& message)
    {
        std::ofstream latest_log(default_logfile, std::ios_base::app);
        latest_log << message;
        latest_log.close();
    }

    void log_to_file(const std::string& filename, const std::string& message)
    {
        std::ofstream latest_log(filename, std::ios_base::app);
        latest_log << message;
        latest_log.close();
    }

    void log_to_stream(std::ostream stream, const std::string& message)
    {
        stream << message;
    }

    std::string toks2msg(std::vector<token> tokens, std::function<std::string(token)> formatter)
    {
        std::string msg;
        for (const token& t: tokens)
        {
            msg += formatter(t);
        }
        return msg;
    }

    std::string toks2consoleMsg(std::vector<token> tokens)
    {
        return toks2msg(tokens, consoleMsgFormatter);
    }

    std::string toks2logfileMsg(std::vector<token> tokens)
    {
        return toks2msg(tokens, logfileMsgFormatter);
    }

    std::string consoleMsgFormatter(token t)
    {
        if (!t.delimiter.empty())
        {
            return std::format("{}{}{}{}{} ", mcolor::toEscapeCode(t.colorCode), t.delimiter[0], t.content, t.delimiter[1], mcolor::toEscapeCode(mcolor::ansiColors::Colors::RESET));
        }

        return std::format("{}{}{} ", mcolor::toEscapeCode(t.colorCode), t.content, mcolor::toEscapeCode(mcolor::ansiColors::Colors::RESET));
    }

    std::string logfileMsgFormatter(token t)
    {
        if (!t.delimiter.empty())
        {
            return std::format("{}{}{} ", t.delimiter[0], t.content, t.delimiter[1]);
        }

        return t.content + " ";
    }

    void log(const std::vector<token>& tokens, const std::string& filename)
    {
        if (console_logging) {
            log_to_console(toks2consoleMsg(tokens));
            log_to_console("\n");
        }
        log_to_file(filename, toks2logfileMsg(tokens));
        log_to_file(filename, "\n");
    }

    void log(const std::vector<token>& tokens)
    {
        log(tokens, default_logfile);
    }

    // Mainly for debug purposes
    void ltlog(const std::vector<token>& tokens)
    {
        std::vector<token> wtokens;
        //auto head = token{.colorCode = ansi_escape_codes::true_color(color_codes::BG_RED.rgb.r, color_codes::BG_RED.rgb.g, color_codes::BG_RED.rgb.b), .content = "GAS"};
        //wtokens.push_back(head);
        wtokens.insert(wtokens.end(), tokens.begin(), tokens.end());

        log(wtokens);
        log_to_file("logtest.log", toks2logfileMsg(wtokens) + "\n");
    }

    std::vector<token> trace_format(
        const std::string& func,
        const std::string& file,
        int line)
    {
        auto trace = token{.content = func};
        auto filedata = token{.content = std::format("{}:{}", file, line)};
        return {trace, filedata};
    }

    std::vector<token> info_format(const std::string& message)
    {
        return log_format("INFO", message);
    }

    std::vector<token> warning_format(const std::string& message)
    {
        return log_format("INFO", message, mcolor::ansiColors::Colors::FG_YELLOW);
    }

    std::vector<token> error_format(const std::string& message)
    {
        return log_format("ERROR", message, mcolor::ansiColors::Colors::FG_RED);
    }

    std::vector<token> fatal_format(const std::string& message)
    {
        return log_format("FATAL", message, mcolor::ansiColors::Colors::FG_BRIGHT_RED);
    }

    std::vector<token> verbose_format(const std::string& message)
    {
        return log_format("VERBOSE", message, mcolor::ansiColors::Colors::FG_CYAN);
    }

    std::vector<token> debug_format(const std::string& message)
    {
        return log_format("DEBUG", message, mcolor::ansiColors::Colors::FG_GREEN);
    }

    std::vector<token> debug_detailed_format(
        const std::string& message,
        const std::string& func,
        const std::string& file,
        int line)
    {
        return log_detailed_format("DEBUG", message, func, file, line, mcolor::ansiColors::Colors::FG_GREEN);
    }

    std::vector<token> info_detailed_format(
        const std::string& message,
        const std::string& func,
        const std::string& file,
        int line)
    {
        return log_detailed_format("INFO", message, func, file, line);
    }

    std::vector<token> warning_detailed_format(
        const std::string& message,
        const std::string& func,
        const std::string& file,
        int line)
    {
        return log_detailed_format("WARNING", message, func, file, line, mcolor::ansiColors::Colors::FG_YELLOW);
    }

    std::vector<token> error_detailed_format(
        const std::string& message,
        const std::string& func,
        const std::string& file,
        int line)
    {
        return log_detailed_format("ERROR", message, func, file, line, mcolor::ansiColors::Colors::FG_RED);
    }

    std::vector<token> verbose_detailed_format(
        const std::string& message,
        const std::string& func,
        const std::string& file,
        int line)
    {
        return log_detailed_format("VERBOSE", message, func, file, line, mcolor::ansiColors::Colors::FG_CYAN);
    }

    std::vector<token> fatal_detailed_format(
        const std::string& message,
        const std::string& func,
        const std::string& file,
        int line)
    {
        return log_detailed_format("FATAL", message, func, file, line, mcolor::ansiColors::Colors::FG_BRIGHT_RED);
    }

    std::vector<token> log_format(
        const std::string& severity_name,
        const std::string& message,
        const mcolor::ansiColors::Colors& severity_cc)
    {
        auto severity = token{.colorCode = severity_cc, .content = severity_name};
        auto content = token{.content = message, .delimiter = ""};
        return {severity, content};
    }

    std::vector<token> log_detailed_format(
        const std::string& severity_name,
        const std::string& message,
        const std::string& func,
        const std::string& file,
        int line,
        const mcolor::ansiColors::Colors& severity_cc)
    {
        std::vector<token> tokens;
        auto timestamp = token{.content = jlog::get_timestamp()};
        auto tf_tokens = jlog::trace_format(func, file, line);
        auto lf_tokens = jlog::log_format(severity_name, message, severity_cc);
        tokens.push_back(timestamp);
        tokens.insert(tokens.end(), tf_tokens.begin(), tf_tokens.end());
        tokens.insert(tokens.end(), lf_tokens.begin(), lf_tokens.end());
        return tokens;
    }
}
